from collections.abc import Mapping
from pathlib import Path

from .openers import from_file


def files(path, extensions=None):
    p = Path(path)
    p = p.expanduser()

    return [
        child
        for child in sorted(p.iterdir())
        if child.is_file and (not extensions or child.suffix in extensions)
    ]


def file_text(fpath):
    path = Path(fpath)
    return path.read_text()


def is_mapping(x):
    return isinstance(x, Mapping)


def from_files_it(files, ignore_unopenable=True):
    for f in files:
        try:
            results = from_file(f)
        except ValueError:
            if ignore_unopenable:
                continue
            raise
        if is_mapping(results):
            for k, result in results.items():
                yield f.with_name(f"{f.name}::{k}"), result
        else:
            yield f, results


def from_files(files, ignore_unopenable=True):
    return dict(from_files_it(files, ignore_unopenable))



from decimal import Decimal
from decimal import InvalidOperation as InvalidOp
from numbers import Number


class AnnotationsMixin:
    def annotate_histogram_amplitudes(self):
        def is_numeric(x):
            return isinstance(x, Number) and not isinstance(x, bool)

        for k in self.keys():
            col = self[k]
            filtered = [_ for _ in col if is_numeric(_)]

            if not filtered:
                continue

            _min = min(filtered)
            _max = max(filtered)
            min0 = min(_min, 0.0)
            max0 = max(0.0, _max)
            _range = Decimal(max0) - Decimal(min0)

            _subclasses_by_type = {}

            for r in self:
                v = r[k]

                if not is_numeric(v):
                    continue

                v0, v1 = sorted([0.0, v])

                try:
                    d = Decimal(v0) - Decimal(min0)
                    start = Decimal(d) / Decimal(_range)
                    h0 = float(start)
                except InvalidOp:
                    h0 = 0.0

                try:
                    d = Decimal(v1) - Decimal(min0)
                    end = Decimal(d) / Decimal(_range)
                    h1 = float(end)
                except InvalidOp:
                    h1 = 0.0

                datatype = type(v)

                if datatype not in _subclasses_by_type:
                    Subclass = type(datatype.__name__.title(), (datatype,), {})
                    _subclasses_by_type[datatype] = Subclass

                Subclass = _subclasses_by_type[datatype]

                replacement = Subclass(v)
                replacement.histo = (h0 * 100, h1 * 100)
                r[k] = replacement



from .uuids import deterministic_uuid


class Result(dict):
    def __init__(self, *args, **kwargs):
        self.annotations = {}
        super().__init__(*args, **kwargs)

    def deterministic_uuid(self, columns, uuid_domain):
        return deterministic_uuid(
            uuid_domain, {k: v for k, v in self.items() if k in columns}
        )

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError

    def __setattr__(self, name, value):
        if name == "annotations":
            object.__setattr__(self, name, value)
        else:
            self[name] = value

    def __getitem__(self, key):
        if isinstance(key, int):
            return list(self.values())[key]
        return super().__getitem__(key)

    def __iter__(self):
        for k, v in self.items():
            if k != "annotations":
                yield v

    def scalar(self):
        return self[0]



import csv
from itertools import islice
from pathlib import Path

from ..resultset import Results
from ..util import is_pathlike

DEFAULT_SNIFF_BYTES = 10 * 1024  # 10kiB


def csv_column_names(path):
    return from_csv(path, max_rows=0).keys()


def sniff_csv_dialect(f, sniff_bytes=None, seek_back_to_start=True):
    sniff_bytes = sniff_bytes or DEFAULT_SNIFF_BYTES
    data = f.read(sniff_bytes)

    if seek_back_to_start:
        f.seek(0)
    sniffer = csv.Sniffer()
    sniffed = sniffer.sniff(data)
    return sniffed


def csv_raw_rows_it(f, dialect=None, sniff=False, *args, **kwargs):
    dialect = dialect
    if not dialect and sniff:
        dialect = sniff_csv_dialect(f)
    return csv.DictReader(f, dialect=dialect, *args, **kwargs)


def csv_rows_it(f, renamed_keys=None, *args, **kwargs):
    for row in csv_raw_rows_it(f):
        if renamed_keys:
            row = dict(zip(renamed_keys, row.values()))
        yield row


def from_csv(
    f, *args, dialect=None, encoding=None, max_rows=None, sniff=False, **kwargs
):
    def make_results(reader):
        fieldnames = reader.fieldnames

        if max_rows is not None:
            reader = islice(reader, max_rows)

        r = Results(reader)
        r._keys_if_empty = fieldnames
        return r

    if is_pathlike(f):
        encoding = encoding or "utf-8-sig"

        with Path(f).open(encoding=encoding) as f:
            rows_iterator = csv_raw_rows_it(
                f, *args, dialect=dialect, sniff=sniff, **kwargs
            )
            return make_results(rows_iterator)
    else:
        rows_iterator = csv_raw_rows_it(
            f, *args, dialect=dialect, sniff=sniff, **kwargs
        )
        return make_results(rows_iterator)


def from_psv(f, *args, dialect=None, **kwargs):
    class PSVDialect(csv.excel):
        delimiter = "|"

    dialect = dialect or PSVDialect()
    return from_csv(f, *args, dialect=dialect, **kwargs)


def from_tsv(f, *args, dialect=None, **kwargs):
    dialect = dialect or csv.excel_tab()
    return from_csv(f, *args, dialect=dialect, **kwargs)


def write_csv_to_filehandle(f, rows, **kwargs):
    try:
        first_row = rows[0]
        rows = rows[1:]
    except TypeError:
        first_row = next(rows)

    fieldnames = list(first_row.keys())

    class Dialect(csv.excel):
        lineterminator = "\n"

    w = csv.DictWriter(f, fieldnames=fieldnames, dialect=Dialect(), **kwargs)
    w.writeheader()
    w.writerow(first_row)
    w.writerows(rows)


def write_csv_to_f(f, rows, encoding=None):
    if is_pathlike(f):
        with Path(f).expanduser().open("w", encoding=encoding, newline="") as _f:
            write_csv_to_filehandle(_f, rows)
    else:
        write_csv_to_filehandle(f, rows)



import io
import itertools
from itertools import groupby
from numbers import Number

from .annotations import AnnotationsMixin
from .cleaning import standardized_key_mapping
from .result import Result
from .saving import save_xlsx
from .sqlutil import create_table_statement
from .typeguess import guess_sql_column_type
from .util import noneslarger as _noneslarger
from .util import nonessmaller as _nonessmaller


def results(rows):
    return Results(rows)


def resultproxy_to_results(rp):
    if rp.returns_rows:
        cols = rp.context.cursor.description
        keys = [c[0] for c in cols]

        r = Results(rp)
        r._keys_if_empty = keys
        return r
    else:
        return None


def get_keyfunc(column=None, columns=None, noneslarger=True):
    if noneslarger is True:
        f = _noneslarger
    elif noneslarger is False:
        f = _nonessmaller
    elif noneslarger is None:

        def ident(x):
            return x

        f = ident

    def _keyfunc(x):
        if column:
            return f(x[column])
        if columns:
            return tuple(f(x[k]) for k in columns)

    return _keyfunc


class Results(list, AnnotationsMixin):
    def __init__(self, *args, **kwargs):
        try:
            given = args[0]
            given = [Result(_) for _ in given]

            args = list(args)
            args[0] = given
            args = tuple(args)

            self.paging = None
        except IndexError:
            pass
        self._keys_if_empty = None
        super().__init__(*args, **kwargs)

    def with_join(
        self,
        other,
        column=None,
        columns=None,
        left=False,
        right=False,
        noneslarger=True,
    ):

        a = self
        b = other

        a_keys = self.keys()
        b_keys = other.keys()

        a_other = {_: None for _ in a_keys if _ not in b_keys}
        b_other = {_: None for _ in b_keys if _ not in a_keys}

        if column is None and columns is None:
            column = a_keys[0]

        keyfunc = get_keyfunc(column, columns, noneslarger=noneslarger)

        b_grouped = b.grouped_by(
            column=column, columns=columns, noneslarger=noneslarger
        )

        if right:
            a_grouped = a.grouped_by(
                column=column, columns=columns, noneslarger=noneslarger
            )

        def do_join_it():
            for a_row in a:
                k = keyfunc(a_row)

                if noneslarger is not None:
                    k = k[1]

                if k in b_grouped and k is not None:
                    for b_row in b_grouped[k]:
                        yield {**a_row, **b_row}
                elif left:
                    yield {**a_row, **b_other}

            if right:
                for b_row in b:
                    k = keyfunc(b_row)

                    if noneslarger is not None:
                        k = k[1]

                    if k is None or k not in a_grouped:
                        yield {**a_other, **b_row}

        return Results(do_join_it())

    def transposed(self):
        first_key, *remainder = self.keys()

        first_values = self[first_key]

        new_keys = (first_key, *first_values)

        def do_it():
            for k in remainder:
                yield dict(zip(new_keys, [k, *(self[k])]))

        return Results(do_it())

    def all_keys(self):
        keylist = dict()

        for row in self:
            rowkeys = row.keys()

            for key in rowkeys:
                if key not in keylist:
                    keylist[key] = True
        return list(keylist.keys())

    def by_key(self, key, value=None):
        def get_value(row):
            if value is None:
                return row
            else:
                return row[value]

        return {_[key]: get_value(_) for _ in self}

    def with_key_superset(self):
        all_keys = self.all_keys()

        def dict_with_all_keys(d):
            return {k: d.get(k, None) for k in all_keys}

        return Results([dict_with_all_keys(_) for _ in self])

    def with_renamed_keys(
        self,
        mapping,
        keep_unmapped_keys=True,
        fail_on_unmapped_keys=False,
        fail_on_overwrite=True,
    ):
        def renamed_key(x):
            try:
                return mapping[x]
            except KeyError:
                if fail_on_unmapped_keys:
                    raise ValueError(f"unmapped key: {x}")
                if keep_unmapped_keys:
                    return x

        def renamed_it():
            for row in self:
                d = {
                    renamed_key(k): v
                    for k, v in row.items()
                    if renamed_key(k) is not None  # noqa
                }
                yield d

        # overwrite_check

        if fail_on_overwrite:
            original = self.keys()
            k_dict = {k: None for k in original}

            for o in original:
                renamed = renamed_key(o)

                if renamed != o:
                    if renamed in k_dict:
                        raise ValueError(
                            f"renaming {o} to {renamed} would overwrite {renamed}"
                        )

                k_dict.pop(o)
                if renamed is not None:
                    k_dict[renamed] = None

        return Results(renamed_it())

    def standardized_key_mapping(self):
        return standardized_key_mapping(self.keys())

    def with_standardized_keys(self):
        return self.with_renamed_keys(self.standardized_key_mapping())

    def with_reordered_keys(
        self, ordering, include_nonexistent=False, include_unordered=False
    ):
        if include_unordered:
            oset = set(ordering)
            ordering = ordering + [k for k in self.keys() if k not in oset]

        return Results(
            {k: _.get(k) for k in ordering if k in _ or include_nonexistent}
            for _ in self
        )

    def strip_values(self):
        for row in self:
            for k, v in row.items():
                if v and isinstance(v, str):
                    stripped = v.strip()

                    if stripped != v:
                        row[k] = stripped

    def strip_all_values(self):
        self.strip_values()

    def standardize_spaces(self):
        self.clean_whitespace()

    def clean_whitespace(self):
        for row in self:
            for k, v in row.items():
                if v and isinstance(v, str):
                    standardized = " ".join(v.split())

                    if standardized != v:
                        row[k] = standardized

    def delete_key(self, column=None):
        self.delete_keys([column])

    def delete_keys(self, columns=None):
        if not self:
            if self._keys_if_empty:
                for column in columns:
                    if column in self._keys_if_empty:
                        self._keys_if_empty.remove(column)
            return

        for row in self:
            for c in columns:
                try:
                    del row[c]
                except KeyError:
                    pass

    def set_blanks_to_none(self):
        for row in self:
            for k, v in row.items():
                if isinstance(v, str) and not v.strip():
                    row[k] = None

    def replace_values(self, before, after):
        for row in self:
            for k, v in row.items():
                if v == before:
                    row[k] = after

    def values_for(self, column=None, columns=None):
        if column is not None:
            values = [_[column] for _ in self]
        elif columns is not None:
            values = [tuple(_[c] for c in columns) for _ in self]
        else:
            values = list(self.values())
        return values

    def pop(self, column=None, default=None):
        if column is None:
            return list.pop(self)

        if isinstance(column, int):
            return list.pop(self, column)

        values = self.values_for(column=column)

        if column:
            columns = [column]
        self.delete_keys(columns)

        return values

    def distinct_values(self, column=None, columns=None):
        values = self.values_for(column, columns)

        d = {k: True for k in values}
        return list(d.keys())

    @property
    def csv(self):
        from .openers import write_csv_to_filehandle

        f = io.StringIO()
        write_csv_to_filehandle(f, self)
        return f.getvalue()

    def save_csv(self, destination):
        from .openers import write_csv_to_f

        write_csv_to_f(destination, self)

    def save_xlsx(self, destination):
        save_xlsx({"Sheet1": self}, destination)

    def keys(self):
        try:
            first = self[0]
        except IndexError:
            if self._keys_if_empty is None:
                return []
            else:
                return self._keys_if_empty
        return list(first.keys())

    def copy(self):
        return Results(self)

    def grouped_by(self, column=None, columns=None, sort=True, noneslarger=True):
        keyfunc = get_keyfunc(column, columns, noneslarger=noneslarger)

        copied = Results(self)

        if sort:
            copied.sort(key=keyfunc)

        def grouped_by_it():
            for k, g in itertools.groupby(copied, key=keyfunc):
                if noneslarger is not None:
                    if columns:
                        k = tuple(_[1] for _ in k)
                    else:
                        k = k[1]

                yield k, Results(g)

        return dict(grouped_by_it())

    def ltrsort(self, key=None, reverse=None, noneslarger=True, **kwargs):
        rev_keys = self.keys()

        if isinstance(reverse, bool):
            sort_order = {_: reverse for _ in rev_keys}
        else:
            reverse = reverse or []
            reverse = set(reverse)
            sort_order = {_: _ in reverse for _ in rev_keys}

        # sort from last -> first column
        sort_order = dict(reversed(sort_order.items()))

        for k, g in groupby(sort_order.items(), key=lambda x: x[1]):
            g = list(g)

            cols = [_[0] for _ in reversed(g)]

            keyfunc = get_keyfunc(column=None, columns=cols, noneslarger=noneslarger)

            self.sort(reverse=k, key=keyfunc)

    def __getitem__(self, x):
        if isinstance(x, slice):
            return Results(list(self)[x])
        elif isinstance(x, Number):
            return list.__getitem__(self, x)
        else:
            return [_[x] for _ in self]

    def one(self):
        length = len(self)
        if not length:
            raise RuntimeError("should be exactly one result, but there is none")
        elif length > 1:
            raise RuntimeError("should be exactly one result, but there is multiple")
        return self[0]

    def scalar(self):
        return self.one()[0]

    def pivoted(self):
        from .pivoting import pivoted

        return pivoted(self)
        try:
            down, across, values = self.keys()
        except ValueError:
            raise ValueError("pivoting requires exactly 3 columns")

        downvalues = self.distinct_values(down)
        acrossvalues = self.distinct_values(across)

        d = {(row[down], row[across]): row[values] for row in self}

        def pivoted_it():
            for downvalue in downvalues:
                out = {down: downvalue}
                row = {
                    acrossvalue: d.get((downvalue, acrossvalue), None)
                    for acrossvalue in acrossvalues
                }
                out.update(row)
                yield out

        return Results(pivoted_it())

    def make_hierarchical(self):
        previous = None

        for r in self:
            original = Result(r)
            if previous:
                for k, v in r.items():
                    if previous[k] == v:
                        r[k] = ""
                    else:
                        break
            previous = original

    @property
    def md(self):
        from tabulate import tabulate

        return tabulate(self, headers="keys", tablefmt="pipe")

    def guessed_sql_column_types(self):
        return {k: guess_sql_column_type(self.values_for(k)) for k in self.keys()}

    def guessed_create_table_statement(self, name):
        guessed = self.guessed_sql_column_types()
        return create_table_statement(name, guessed)

    def new_row(self):
        return Result({k: None for k in self.keys()})

    def append_empty_row(self):
        row = self.new_row()
        self.append(row)
        return row



import io
from itertools import islice
from pathlib import Path

import chardet

from ..resultset import Results
from .csv import (  # noqa
    csv_column_names,
    csv_rows_it,
    from_csv,
    from_psv,
    from_tsv,
    sniff_csv_dialect,
    write_csv_to_f,
    write_csv_to_filehandle,
)


def detect_string_enc(contents):
    b = io.BytesIO(contents)
    b.seek(0)
    return detect_enc(b)


def detect_enc(f):
    return chardet.detect(f.read())["encoding"]


def smart_open(f):
    try:
        return io.open(f)
    except TypeError:
        return f


def first_n_lines(stream, n=10):
    return list(islice(stream, n))


def dicts_from_rows(rows):
    try:
        first = rows[0]
    except IndexError:
        return []
    rest = rows[1:]

    def it():
        for d in rest:
            dd = dict(zip(first, d))
            yield dd

    return list(it())


def from_xlsx(f, max_rows=None):
    from openpyxl import load_workbook

    if isinstance(f, Path):
        f = str(f)
    wb = load_workbook(filename=f, read_only=True, data_only=True)

    wsheets = [ws for ws in wb]

    def cell_value(c):
        try:
            internal_value = c.internal_value
            return internal_value or ""
        except AttributeError:
            return ""

    def xget_row_values(row):
        return [cell_value(c) for c in list(row)]

    def do_sheet(ws):
        if max_rows is None:
            rows = list(ws.rows)
        else:
            rows = list(islice(ws.rows, max_rows))

        rows = [xget_row_values(_) for _ in list(rows)]
        return dicts_from_rows(rows)

    return dict(zip(wb.sheetnames, (Results(do_sheet(_)) for _ in wsheets)))


def from_xls(f, file_contents=None):
    from xlrd import open_workbook

    wb = open_workbook(str(f), file_contents=file_contents)

    def get_row_values(ws, rownum):
        return [str(ws.cell_value(rowx=rownum, colx=c) or "") for c in range(ws.ncols)]

    def do_sheet(ws):
        rows = [get_row_values(ws, rx) for rx in range(ws.nrows)]
        return dicts_from_rows(rows)

    wsheets = [wb.sheet_by_index(_) for _ in range(wb.nsheets)]

    return dict(zip(wb.sheet_names(), (Results(do_sheet(_)) for _ in wsheets)))


OPENERS = {
    ".xlsx": from_xlsx,
    ".xls": from_xls,
    ".csv": from_csv,
    ".tsv": from_tsv,
    ".psv": from_psv,
}


def from_file(f, **kwargs):
    p = Path(f)

    extension = p.suffix

    try:
        return OPENERS[extension](f, **kwargs)
    except KeyError:
        if kwargs.get("sniff"):
            return from_csv(f, **kwargs)
        raise ValueError(f"cannot open a file with extension: {extension}")



import base64
import csv
import io
import json
from itertools import zip_longest

from .sqlutil import ordering_from_parsed, parse_ordering, reversed_ordering

sio = io.StringIO
csvreader = csv.reader
csvwriter = csv.writer


PAGED_QUERY = """
select * from
(
{q}
) unpaged_table
{bookmark}
{order_by}
{limit}
"""

PAGED_QUERY_MSSQL = """
select {limit} * from
(
{q}
) unpaged_table
{bookmark}
{order_by}
"""

PARAM_PREFIX = "paging_bookmark_"


def encode_bookmark_values(decoded):
    j = json.dumps(decoded)
    j8 = j.encode("utf-8")
    return base64.urlsafe_b64encode(j8).decode("utf-8")


def decode_bookmark(encoded):
    j = base64.urlsafe_b64decode(encoded)
    jt = j.decode("utf-8")
    return json.loads(jt)


def bind_pairs_iter(cols, bookmark, swap_on_descending=False):
    for i, zipped in enumerate(zip(cols, bookmark)):
        col, val = zipped
        name, is_descending = col
        lowercase_name = name.lower()
        bind = f":{PARAM_PREFIX}{lowercase_name}"

        if swap_on_descending and is_descending:
            yield name, bind
        else:
            yield bind, name


def generate_or_statement(a, b, a_compared, b_compared):
    if a and b:
        a_first, b_first = a[0], b[0]
        or_statement = generate_or_statement(
            a[1:], b[1:], a_compared + [a[0]], b_compared + [b[0]]
        )
        equalities = [f"{x} = {y}" for x, y in zip(a_compared, b_compared)]
        equalities_joined = " and ".join(equalities)
        if or_statement:
            return f"({equalities_joined} and {a_first} > {b_first}) or {or_statement}"
        else:
            return f"({equalities_joined} and {a_first} > {b_first})"


def recursive_comparison(a, b):
    or_statement = generate_or_statement(a[1:], b[1:], [a[0]], [b[0]])
    return f"({a[0]} > {b[0]} or {or_statement})"


def make_bookmark_where_clause(cols, bookmark, backwards=False, supports_row=True):
    if bookmark is None:
        return ""

    pairslist = bind_pairs_iter(cols, bookmark, swap_on_descending=True)

    b, a = zip(*pairslist)
    if len(a) > 1 or len(b) > 1:
        if supports_row:
            a, b = ", ".join(a), ", ".join(b)
            return f"where row({a}) > row({b})"
        else:
            return f"where {recursive_comparison(a, b)}"
    else:
        return f"where {a[0]} > {b[0]}"


def paging_params(cols, bookmark):
    names = [PARAM_PREFIX + c[0].lower() for c in cols]
    return dict(zip_longest(names, bookmark or []))


def paging_wrapped_query(
    query, ordering, bookmark, per_page, backwards, use_top=False, supports_row=True
):
    cols = parse_ordering(ordering)
    if backwards:
        cols = reversed_ordering(cols)

    bookmark_clause = make_bookmark_where_clause(
        cols, bookmark, backwards, supports_row=supports_row
    )
    order_list = ordering_from_parsed(cols)
    order_by = f"order by {order_list}"
    if use_top:
        limit = f"top {per_page + 1}"
    else:
        limit = f"limit {per_page + 1}"

    params = paging_params(cols, bookmark)
    formatted = (PAGED_QUERY_MSSQL if use_top else PAGED_QUERY).format(
        q=query, bookmark=bookmark_clause, order_by=order_by, limit=limit
    )
    return formatted, params


class Paging:
    def __init__(self, results, per_page, ordering, bookmark, backwards):
        self.results = results
        self.per_page = per_page
        self.ordering = ordering
        self.bookmark = bookmark
        self.backwards = backwards
        self.parsed_ordering = parse_ordering(ordering)
        self.order_keys = [c[0] for c in self.parsed_ordering]

        try:
            self.discarded_item = results.pop(per_page)
            self.has_more = True
        except IndexError:
            self.discarded_item = None
            self.has_more = False

        if backwards:
            results.reverse()
        self.count = len(self.results)

    @property
    def has_after(self):
        return self.discarded_item is not None

    @property
    def has_before(self):
        return self.bookmark is not None

    @property
    def has_next(self):
        if self.backwards:
            return self.has_before
        else:
            return self.has_after

    @property
    def has_prev(self):
        if self.backwards:
            return self.has_after
        else:
            return self.has_before

    @property
    def at_start(self):
        return not self.has_prev

    @property
    def at_end(self):
        return not self.has_next

    @property
    def is_all(self):
        return self.at_start and self.at_end

    @property
    def is_full(self):
        return self.count == self.per_page

    @property
    def past_start(self):
        return self.backwards and self.at_start and not self.is_full

    @property
    def past_end(self):
        return not self.backwards and self.at_end and not self.is_full

    @property
    def next(self):
        if not self.is_empty:
            return self.get_bookmark(self.results[-1])

    @property
    def prev(self):
        if not self.is_empty:
            return self.get_bookmark(self.results[0])

    @property
    def is_empty(self):
        return not bool(self.count)

    @property
    def current(self):
        return self.bookmark

    @property
    def current_reversed(self):
        if self.discarded_item:
            return self.get_bookmark(self.discarded_item)

    def get_bookmark(self, result_row):
        return tuple(result_row[k] for k in self.order_keys)



import itertools

from .resultset import Results


class item:
    def __init__(self, i):
        self.i = i

    @property
    def after(self):
        aft = []

        x = self.next

        while x and x.i != self.i and x.i not in aft:
            aft.append(x.i)
            x = x.next

        return aft

    @property
    def before(self):
        bef = []

        x = self.prev

        while x and x.i != self.i and x.i not in bef:
            bef.append(x.i)
            x = x.next

        return bef


class items:
    def __init__(self, _list):

        full = []

        previous = None

        for x in _list:
            i = item(x)

            i.prev = previous

            if previous:
                previous.next = i

            i.next = None
            full.append(i)
            previous = i
        self._list = full


def ordering(groups):
    glist = [items(_) for _ in groups]
    allitems = itertools.chain(*groups)

    d_all = {k: None for k in allitems}

    d_before = {k: set() for k in d_all.keys()}

    for g in glist:
        for i in g._list:
            val = i.i
            d_before[val] = d_before[val] | set(i.before)

    _result = []

    all_keys = {k: None for k in d_all.keys()}

    while all_keys:
        for k in all_keys:
            if d_before[k] - set(_result):
                continue
            _result.append(k)
            all_keys.pop(k)
            break
        else:
            k = list(all_keys.keys())[0]
            _result.append(k)
            all_keys.pop(k)

    return _result


def pivoted(_results):
    r = _results
    try:
        *down, across, values = r.keys()
    except ValueError:
        raise ValueError(
            "Pivoting requires at least 2 columns for input: 1 or more row labels, 0 or 1 column labels, and one values column"
        )

    if not down:
        return Results(_results)

    downvalues = r.distinct_values(columns=down)

    orders = [
        _.distinct_values(column=across) for _ in r.grouped_by(columns=down).values()
    ]
    acrossvalues = ordering(orders)
    key_cols = down + [across]

    def cell_key(row):
        return tuple(row[k] for k in key_cols)

    d = {cell_key(row): row[values] for row in r}

    def pivoted_it():
        for downvalue in downvalues:
            out = {k: v for k, v in zip(down, downvalue)}

            row = {
                acrossvalue: d.get(tuple(list(downvalue) + [acrossvalue]), None)
                for acrossvalue in acrossvalues
            }
            out.update(row)
            yield out

    return Results(pivoted_it())



from pathlib import Path


def is_pathlike(f):
    # return true if this is a pathlib or py.path object
    return isinstance(f, (str, Path)) or hasattr(f, 'read_text')


def noneslarger(x):
    if x is None:
        return 1, x
    else:
        return 0, x


def nonessmaller(x):
    if x is None:
        return -1, x
    else:
        return 0, x


ELLIPSIS = "\u2026"


def trunc(s, length=25):
    if len(s) > length:
        s = s[:length] + ELLIPSIS
    return s



from logx import log
from migra import Migration
from sqlalchemy.pool import NullPool
from sqlbag import create_database  # noqa
from sqlbag import S, temporary_database

from . import connections


class SchemaManagement:
    def create_database(self, *args, **kwargs):
        create_database(self.db_url, *args, **kwargs)

    def has_same_structure_as(self, other_db_url, **kwargs):
        return db_structures_are_identical(self.db_url, other_db_url, **kwargs)

    def creation_statements(self, **kwargs):
        return creation_statements(self.db_url)

    def drop_statements(self, **kwargs):
        return drop_statements(self.db_url)

    def schema_hash(self):
        return schema_hash(self.db_url)

    def db_differences(self, other_db_url, **kwargs):
        return db_differences(self.db_url, other_db_url, **kwargs)

    def sync_db_structure_to_target_db(
        self, target_db_url, confirm=True, create_extensions_only=False, **kwargs
    ):
        return sync_db_structure_to_target_db(
            self.db_url,
            target_db_url,
            confirm=confirm,
            create_extensions_only=create_extensions_only,
            **kwargs,
        )

    def sync_db_structure_to_setup_method(
        self, setup_method, confirm=True, create_extensions_only=False, **kwargs
    ):
        sync_db_structure_to_setup_method(
            self.db_url,
            setup_method,
            confirm=confirm,
            create_extensions_only=create_extensions_only,
            **kwargs,
        )

    def sync_db_structure_to_definition(
        self, definition_text, confirm=True, create_extensions_only=False, **kwargs
    ):
        sync_db_structure_to_definition(
            self.db_url,
            definition_text,
            confirm=confirm,
            create_extensions_only=create_extensions_only,
            **kwargs,
        )


def prompt(question, prompt=True):  # pragma: no cover
    print(question + " ", end="")
    return input().strip().lower() == "y"


def db_structures_are_identical(db_url_a, db_url_b, **kwargs):
    """
    Do two databases have identical schema structure? This method tells you the answer.

    - db_url_a: first database
    - db_url_b: second database
    - **kwargs: additional args, for passing into a `migra` migration object.

    """
    differences = db_differences(db_url_a, db_url_b, **kwargs)
    return not differences.strip()


def creation_statements(db_url, **kwargs):
    with S(db_url, poolclass=NullPool) as s:
        m = Migration(None, s, **kwargs)
        m.set_safety(False)
        m.add_all_changes()
        return m.sql


def schema_hash(db_url, **kwargs):
    import hashlib

    sql = creation_statements(db_url, **kwargs)
    encoded = sql.encode("utf-8")
    hash = hashlib.sha1(encoded)
    return hash.hexdigest()


def drop_statements(db_url, **kwargs):
    with S(db_url, poolclass=NullPool) as s:
        m = Migration(s, None, **kwargs)
        m.set_safety(False)
        m.add_all_changes()
        return m.sql


def db_differences(db_url_a, db_url_b, **kwargs):
    with S(db_url_a, poolclass=NullPool) as a, S(db_url_b, poolclass=NullPool) as b:
        m = Migration(a, b, **kwargs)
        m.set_safety(False)
        m.add_all_changes()
        return m.sql


def sync_db_structure_to_target_db(
    db_url, target_db_url, confirm=True, create_extensions_only=False, **kwargs
):
    log.info(f"syncing: {db_url} to {target_db_url}")

    with S(db_url, poolclass=NullPool) as s_current, S(
        target_db_url, poolclass=NullPool
    ) as s_target:
        m = Migration(s_current, s_target, **kwargs)
        m.set_safety(False)
        if create_extensions_only:
            log.info("Syncing extension creation only...")
            m.add_extension_changes(creates=True, drops=False)
        else:
            m.add_all_changes()

        if m.statements:
            if confirm:  # pragma: no cover
                print("THE FOLLOWING CHANGES ARE PENDING:", end="\n\n")
                print(m.sql)
                print()
            if not confirm or prompt("Apply these changes?"):
                log.info("Applying...")
                m.apply()
                log.info("Applied.")
            else:
                if confirm:  # pragma: no cover
                    print("Not applying.")
        else:
            if confirm:  # pragma: no cover
                print("Already synced.")

    current_schema_hash = schema_hash(db_url)
    if confirm:  # pragma: no cover
        print(f"Schema hash: {current_schema_hash}")


def sync_db_structure_to_setup_method(
    db_url, setup_method, confirm=True, create_extensions_only=False, **kwargs
):
    with temporary_database(host="localhost") as temp_db_url:
        setup_method(temp_db_url)
        sync_db_structure_to_target_db(
            db_url,
            temp_db_url,
            confirm=confirm,
            create_extensions_only=create_extensions_only,
            **kwargs,
        )


def sync_db_structure_to_definition(
    db_url, definition_text, confirm=True, create_extensions_only=False, **kwargs
):
    def load_def(db_url):
        connections.db(db_url, poolclass=NullPool).raw(definition_text)

    sync_db_structure_to_setup_method(
        db_url,
        load_def,
        confirm=confirm,
        create_extensions_only=create_extensions_only,
        **kwargs,
    )



import string

ALLOWED = string.ascii_lowercase + "_" + string.digits


allowed = {k: True for k in ALLOWED}


def whitespace_to_underscore(x):
    return "_".join(x.split())


def standardize_key(k):
    cleaned = k.strip().lower().encode("ascii", errors="ignore").decode()
    underscored = whitespace_to_underscore(cleaned)
    return filtered(underscored)


def filtered(x):
    return "".join(_ for _ in x if _ in allowed)


def standardized_key_mapping(keyslist):
    return {k: standardize_key(k) for k in keyslist}



import json
import uuid


def dns_uuid(domain_name):
    return uuid.uuid5(uuid.NAMESPACE_DNS, domain_name)


def stringified_dict(d):
    return {str(k): str(v) for k, v in d.items()}


def contents_json(contents_dict):
    d = stringified_dict(contents_dict)
    return json.dumps(d, sort_keys=True)


def deterministic_uuid(domain, contents_dict):
    dns = dns_uuid(domain)
    j = contents_json(contents_dict)
    return uuid.uuid5(dns, j)
